//===========================================================================
//
//  Code written for Seeeduino v4.2 runnung at 3.3v
//
//  CRYSTALFONTZ CFAL6448A-066B 64x48 OLED in I2C mode
//
//  ref: https://www.crystalfontz.com/product/cfal6448a066bw
//
//  2017 - 06 - 23 Brent A. Crosby
//===========================================================================
//This is free and unencumbered software released into the public domain.
//
//Anyone is free to copy, modify, publish, use, compile, sell, or
//distribute this software, either in source code form or as a compiled
//binary, for any purpose, commercial or non-commercial, and by any
//means.
//
//In jurisdictions that recognize copyright laws, the author or authors
//of this software dedicate any and all copyright interest in the
//software to the public domain. We make this dedication for the benefit
//of the public at large and to the detriment of our heirs and
//successors. We intend this dedication to be an overt act of
//relinquishment in perpetuity of all present and future rights to this
//software under copyright law.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
//OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
//ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
//OTHER DEALINGS IN THE SOFTWARE.
//
//For more information, please refer to <http://unlicense.org/>
//===========================================================================
#include <Wire.h>
#include "bitmaps.h"
//============================================================================
/*

The CFAL6448A-066B is a 3.3v device. You need a 3.3v Arduino to operate this
code properly. We used a seeedunio v4.2 set to 3.3v:
https://www.crystalfontz.com/product/cfapn15062-seeeduino-arduino-clone-microprocessor
or
http://www.seeedstudio.com/item_detail.html?p_id=2517

Connections:

       Adruino => OLED
Analog 5 / SCL => D0 (add ~5K pullup)
Analog 4 / SDA => D1 and D2 (add ~5K pullup)
        Port 4 => *RESET
           GND => D/C (Sets I2C address)
           GND => BS0 (Selects I2C)
          3.3v => BS1 (Selects I2C)
          3.3v => VBAT
          3.3v => VDD
*/
//============================================================================
#define OLED_RESET 4
#define CLR_RESET (digitalWrite(OLED_RESET, LOW))
#define SET_RESET (digitalWrite(OLED_RESET, HIGH))
//============================================================================
#define NUMBER_OF_SCREENS (13)
#define GEAR_START_SCREEN (5)
const SCREEN_IMAGE *const screens[NUMBER_OF_SCREENS] PROGMEM=
  {
  &Logo_6448,
  &Dimensions_6448,
  &Left_Double_6448,
  &Left_6448,
  &OLED_6448,
  &Gear_0_6448,
  &Gear_1_6448,
  &Gear_2_6448,
  &Gear_3_6448,
  &Gear_4_6448,
  &Gear_5_6448,
  &Gear_6_6448,
  &Gear_7_6448
  };
//============================================================================
void I2C_sendCommand(uint8_t payload)
  {
  Wire.beginTransmission(0x3C); //send start & Slave address
  Wire.write(0x00);		        //Control Byte - Command
  Wire.write(payload);	        //payload
  Wire.endTransmission();
  }
//============================================================================
void I2C_sendData(uint8_t payload)
  {
  Wire.beginTransmission(0x3C);	//send start & Slave address
  Wire.write(0xC0);		        //Control Byte - Data (non-continued)
  Wire.write(payload);          //payload
  Wire.endTransmission();
  }
//============================================================================
#define SSD1306B_00_SET_LOWER_COLUMN_ADDRESS_BIT (0x00)
#define SSD1306B_10_SET_UPPER_COLUMN_ADDRESS_BIT (0x10)
#define SSD1306B_B0_SET_PAGE_START_ADDRESS_BIT   (0xB0)
#define SSD1306B_DCDC_CONFIG_PREFIX_8D           (0x8D)
#define SSD1306B_DCDC_CONFIG_7p5v_14             (0x14)
#define SSD1306B_DCDC_CONFIG_6p0v_15             (0x15)
#define SSD1306B_DCDC_CONFIG_8p5v_94             (0x94)
#define SSD1306B_DCDC_CONFIG_9p0v_95             (0x95)
#define SSD1306B_DISPLAY_OFF_YES_SLEEP_AE        (0xAE)
#define SSD1306B_DISPLAY_ON_NO_SLEEP_AF          (0xAF)
#define SSD1306B_CLOCK_DIVIDE_PREFIX_D5          (0xD5)
#define SSD1306B_MULTIPLEX_RATIO_PREFIX_A8       (0xA8)
#define SSD1306B_DISPLAY_OFFSET_PREFIX_D3        (0xD3)
#define SSD1306B_DISPLAY_START_LINE_40           (0x40)
#define SSD1306B_SEG0_IS_COL_0_A0                (0xA0)
#define SSD1306B_SEG0_IS_COL_127_A1              (0xA1)
#define SSD1306B_SCAN_DIR_UP_C0                  (0xC0)
#define SSD1306B_SCAN_DIR_DOWN_C8                (0xC8)
#define SSD1306B_COM_CONFIG_PREFIX_DA            (0xDA)
#define SSD1306B_COM_CONFIG_SEQUENTIAL_LEFT_02   (0x02)
#define SSD1306B_COM_CONFIG_ALTERNATE_LEFT_12    (0x12)
#define SSD1306B_COM_CONFIG_SEQUENTIAL_RIGHT_22  (0x22)
#define SSD1306B_COM_CONFIG_ALTERNATE_RIGHT_32   (0x32)
#define SSD1306B_CONTRAST_PREFIX_81              (0x81)
#define SSD1306B_PRECHARGE_PERIOD_PREFIX_D9      (0xD9)
#define SSD1306B_VCOMH_DESELECT_PREFIX_DB        (0xDB)
#define SSD1306B_VCOMH_DESELECT_0p65xVCC_00      (0x00)
#define SSD1306B_VCOMH_DESELECT_0p71xVCC_10      (0x10)
#define SSD1306B_VCOMH_DESELECT_0p77xVCC_20      (0x20)
#define SSD1306B_VCOMH_DESELECT_0p83xVCC_30      (0x30)
#define SSD1306B_ENTIRE_DISPLAY_FORCE_ON_A5      (0xA5)
#define SSD1306B_ENTIRE_DISPLAY_NORMAL_A4        (0xA4)
#define SSD1306B_INVERSION_NORMAL_A6             (0xA6)
#define SSD1306B_INVERSION_INVERTED_A7           (0xA7)
//============================================================================
void Set_Address(uint8_t column, uint8_t page)
  {
  //Set column-lower nibble
  I2C_sendCommand(SSD1306B_00_SET_LOWER_COLUMN_ADDRESS_BIT|(column&0x0F));
  //Set column-upper nibble
  I2C_sendCommand(SSD1306B_10_SET_UPPER_COLUMN_ADDRESS_BIT|((column>>4)&0x0F));
  //Set page address, limiting from 0 to 7
  I2C_sendCommand(SSD1306B_B0_SET_PAGE_START_ADDRESS_BIT|(page&0x0F));
  }
//============================================================================
void show_64_x_48_bitmap(const SCREEN_IMAGE *OLED_image)
  {
  // ~ 16.26 mS normal I2C_sendData() calls are ~45 mS
  uint8_t
    column;
  uint8_t
    row;

  for(row=0;row<6;row++)
    {
    //Display is offest by 32 columns
    //(only 64 used of a possible 128, 32 discarded each side)
    Set_Address(32,row);
    //Arduino TWI Library is max 32 bytes by default.
    //Break into a right and left half.
    //Left Half
    Wire.beginTransmission(0x3C);  //send start & Slave address
    Wire.write(0x40);   //Control Byte - Data (continued)
    for(column=0;column<32;column++)
      {
      //Read this byte from the program memory / flash
      Wire.write(pgm_read_byte( &(OLED_image  ->bitmap_data[row][column]) ));
      }
    //Right Half
    Wire.endTransmission();
    Wire.beginTransmission(0x3C);  //send start & Slave address
    Wire.write(0x40);   //Control Byte - Data (continued)
    for(column=32;column<64;column++)
      {
      //Read this byte from the program memory / flash
      Wire.write(pgm_read_byte( &(OLED_image  ->bitmap_data[row][column]) ));
      }
    Wire.endTransmission();
    }
  }
//============================================================================
void Initialize_CFAL6448A(void)
  {
  //Thump the reset.  
  _delay_ms(1);
  CLR_RESET;
  _delay_ms(1);
  SET_RESET;
  _delay_ms(1);

  //Set the display to sleep mode for the rest of the init.
  I2C_sendCommand(SSD1306B_DISPLAY_OFF_YES_SLEEP_AE);

  //Set the clock speed, nominal ~105FPS
  //Low nibble is divide ratio
  //High level is oscillator frequency
  I2C_sendCommand(SSD1306B_CLOCK_DIVIDE_PREFIX_D5);
  I2C_sendCommand(0x80);//177Hz measured

  //Set the multiplex ratio to 1/48
  //Default is 0x3F (1/64 Duty), we need 0x2F (1/48 Duty)
  I2C_sendCommand(SSD1306B_MULTIPLEX_RATIO_PREFIX_A8);
  I2C_sendCommand(0x2F);  

  //Set the display offset to 0 (default)
  I2C_sendCommand(SSD1306B_DISPLAY_OFFSET_PREFIX_D3);
  I2C_sendCommand(0x00);

  //Set the display RAM display start line to 0 (default)
  //Bits 0-5 can be set to 0-63 with a bitwise or
  I2C_sendCommand(SSD1306B_DISPLAY_START_LINE_40);

  //Enable DC/DC converter, 9.0v
  I2C_sendCommand(SSD1306B_DCDC_CONFIG_PREFIX_8D);
  I2C_sendCommand(SSD1306B_DCDC_CONFIG_9p0v_95);

  //Map the columns correctly for our OLED glass layout
  I2C_sendCommand(SSD1306B_SEG0_IS_COL_127_A1);

  //Set COM output scan correctly for our OLED glass layout
  I2C_sendCommand(SSD1306B_SCAN_DIR_DOWN_C8);

  //Set COM pins correctly for our OLED glass layout
  I2C_sendCommand(SSD1306B_COM_CONFIG_PREFIX_DA);
  I2C_sendCommand(SSD1306B_COM_CONFIG_ALTERNATE_LEFT_12);

  //Set Contrast Control / SEG Output Current / Iref
  I2C_sendCommand(SSD1306B_CONTRAST_PREFIX_81);
  I2C_sendCommand(0xFF);  //magic # from factory

  //Set precharge (low nibble) / discharge (high nibble) timing
  //precharge = 15 clocks
  //discharge = 15 clocks
  I2C_sendCommand(SSD1306B_PRECHARGE_PERIOD_PREFIX_D9); //Set Pre-Charge period        
  I2C_sendCommand(0xFF); 

  //Set VCOM Deselect Level
  I2C_sendCommand(SSD1306B_VCOMH_DESELECT_PREFIX_DB);
  I2C_sendCommand(SSD1306B_VCOMH_DESELECT_0p65xVCC_00); 

  //Make sure Entire Display On is disabled (default)
  I2C_sendCommand(SSD1306B_ENTIRE_DISPLAY_NORMAL_A4);

  //Make sure display is not inverted (default)
  I2C_sendCommand(SSD1306B_INVERSION_NORMAL_A6);

  //Get out of sleep mode, into normal operation
  I2C_sendCommand(SSD1306B_DISPLAY_ON_NO_SLEEP_AF);
  }
//============================================================================
uint8_t
  gear_dir;
//----------------------------------------------------------------------------
void setup( void )
  {
  // Initialize digital pin 4 as an output for the reset.
  pinMode(OLED_RESET, OUTPUT);

  //Get the I2C going
  Wire.begin();

  //Thanks to Limor / Adafruit for the hint ;-)   
  TWBR = 12; // upgrade to 400KHz!

  //Fire up the I2C OLED
  Initialize_CFAL6448A();

  gear_dir=0;
  }
//----------------------------------------------------------------------------
void  loop(void)
  {
  uint8_t
    current_screen;
  //Put up some bitmaps from flash
  for(current_screen=0;current_screen<GEAR_START_SCREEN;current_screen++)
    {
    show_64_x_48_bitmap((SCREEN_IMAGE *)pgm_read_word(&screens[current_screen]));
    //Wait a bit . . .
    _delay_ms(1500);
    if(current_screen < 2)
      {
      //Wait a little extra for the logo & dimensions
      _delay_ms(750);
      }
    }
  //Loop the gear animation as fast as the I2C can stuff it out.
  uint8_t
    repeat;

  if(0==gear_dir)
    {
    for(repeat=0;repeat<25;repeat++)
      {
      for(current_screen=GEAR_START_SCREEN;current_screen<NUMBER_OF_SCREENS;current_screen++)
        {
        show_64_x_48_bitmap((SCREEN_IMAGE *)pgm_read_word(&screens[current_screen]));
		//I2C slow enough to not need a delay
        }
      }
    gear_dir=1;  
    }
  else
    {
    for(repeat=0;repeat<25;repeat++)
      {
      for(current_screen=NUMBER_OF_SCREENS-1;GEAR_START_SCREEN<=current_screen;current_screen--)
        {
        show_64_x_48_bitmap((SCREEN_IMAGE *)pgm_read_word(&screens[current_screen]));
		//I2C slow enough to not need a delay
        }
      }
    gear_dir=0;  
    }
  }
//============================================================================

